#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "config.h"
#include "imapfilter.h"
#include "data.h"


extern int sockpri, sockaux;
extern unsigned int options;
extern unsigned int capabilities;

namesp_t nsppri, nspaux;	/* Primary and auxiliary namespace. */


#ifdef DEBUG
/*
 * Test/ping server.
 */
int test(int *sock)
{
    return server_response(sock, imap_noop(sock));
}

#endif


/*
 * Check server's capabilities.
 */
int check_capabilities(int *sock)
{
    capabilities = CAPABILITY_NONE;

    return capability_response(sock, imap_capability(sock));
}


/*
 * Get namespace of mail server's mailboxes.
 */
int check_namespace(int *sock, namesp_t * nsp)
{
    nsp->prefix[0] = nsp->delim = 0;

    if (!(options & OPTION_NAMESPACE) ||
	!(capabilities & CAPABILITY_NAMESPACE))
	return 0;
    else
	return namespace_response(sock, imap_namespace(sock), nsp);
}


/*
 * Login to server.
 */
int login(int *sock, char *user, char *pass)
{
    log_info(LOG_USERNAME, user);

    return server_response(sock, imap_login(sock, user, pass));
}



/*
 * Check if a mailbox exists.
 */
int check_mailbox(int *sock, char *mbox, namesp_t * nsp)
{
    return server_response(sock, imap_examine(sock,
					      apply_namespace(mbox,
							      nsp->prefix,
							      nsp->delim)));
}


/*
 * Open mailbox in read-write mode.
 */
int select_mailbox(int *sock, char *mbox, namesp_t * nsp)
{
    int r;

    if (mailbox_status(sock, mbox, nsp) == -2)
	return -2;		/* No messages exist. No filters need to be
				   applied. */

    r = select_response(sock, imap_select(sock,
					  apply_namespace(mbox, nsp->prefix,
							  nsp->delim)));

    log_info(LOG_MAILBOX, mbox);

    return r;
}


/*
 * Get mailbox's status.
 */
int mailbox_status(int *sock, char *mbox, namesp_t * nsp)
{
    return status_response(sock, imap_status(sock,
					  apply_namespace(mbox, nsp->prefix,
							  nsp->delim),
					   "MESSAGES RECENT UNSEEN"), mbox);
}


/*
 * Close examined/selected mailbox.
 */
int close_mailbox(int *sock)
{
    return server_response(sock, imap_close(sock));
}


/*
 * Logout from server.
 */
int logout(int *sock)
{
    return logout_response(sock, imap_logout(sock));
}


/*
 * Match and apply filters assigned to a mailbox.
 */
int apply_filters(filter_t ** filters)
{
    int i;
    char *mesgs;

    for (i = 0; filters[i]; i++) {
	mesgs = NULL;

	if (match_filter(filters[i], &mesgs))
	    continue;

	log_info(LOG_FILTER, filters[i]->key);

	apply_action(mesgs, &(filters[i]->action.type),
		   filters[i]->action.raccount, filters[i]->action.destmbox,
		     &filters[i]->action.msgflags, filters[i]->action.args);

	xfree(mesgs);
    }

    return 0;
}


/*
 * Generate the search request by the masks of the filter and try to
 * match the generated filter.
 */
int match_filter(filter_t * filter, char **mesgs)
{
    char *search;

    if (filter->mode == FILTER_MODE_OR)
	search = generate_filter_or(filter->masks, filter->masknum,
				    filter->masklen);
    else
	search = generate_filter_and(filter->masks, filter->masknum,
				     filter->masklen);

    search_response(&sockpri, imap_search(&sockpri, search), mesgs);

    xfree(search);

    if (!*mesgs)
	return 1;

    return 0;
}


/*
 * Empty the FIFO inventory.
 */
void empty_fifo(mask_t ** mfifo)
{
    mfifo[0] = NULL;

    queue_fifo(NULL, NULL);
    dequeue_fifo(NULL);
}


/*
 * Add item to FIFO inventory.
 */
void queue_fifo(mask_t ** mfifo, mask_t * mask)
{
    static unsigned int i;

    if (!mfifo) {
	i = 0;
	return;
    }
    mfifo[i++] = mask;
    mfifo[i] = NULL;
}


/*
 * Get next item from FIFO inventory.
 */
mask_t *dequeue_fifo(mask_t ** mfifo)
{
    static unsigned int j;

    if (!mfifo) {
	j = 0;
	return NULL;
    }
    return mfifo[j++];
}


/*
 * Generate the filter search command from the masks, assuming that
 * masks are AND-ed.
 */
char *generate_filter_and(mask_t * mask, unsigned int masknum,
			   unsigned int masklen)
{
    const unsigned int searchbuf = masklen + masknum * 6 + 8;
    unsigned int len = 0;
    char *search;
    mask_t *tmp;

    search = (char *)xmalloc(sizeof(char) * searchbuf);

    search[0] = 0;

    tmp = mask;
    if (!tmp) {
	strncat(search, "ALL ", searchbuf - len - 1);
	len += 4;
    } else
	while ((tmp = tmp->next)) {
	    if (tmp->type != MASK_TYPE_OR) {
		strncat(search, "ALL ", searchbuf - len - 1);
		len += 4;
		break;
	    }
	}

    tmp = NULL;
    while (mask) {
	tmp = mask;
	mask = mask->next;

	if (mask && mask->type == MASK_TYPE_OR) {
	    strncat(search, "OR (", searchbuf - len - 1);
	    len += 4;

	    strncat(search, tmp->body, searchbuf - len - 1);
	    len = strlen(search);
	    search[len] = ' ';
	    search[++len] = 0;

	    search[len - 1] = ')';
	    search[len] = ' ';
	    search[++len] = 0;

	    if (!mask->next || mask->next->type != MASK_TYPE_OR) {
		search[len] = '(';
		search[++len] = 0;
		strncat(search, mask->body, searchbuf - len - 1);
		len = strlen(search);
		search[len] = ')';
		search[++len] = ' ';
		search[++len] = 0;
		mask = mask->next;
	    }
	} else {
	    strncat(search, tmp->body, searchbuf - len - 1);
	    len = strlen(search);
	    search[len] = ' ';
	    search[++len] = 0;
	}
    }

    search[len - 1] = 0;

    return search;
}


/*
 * Generate the filter search command from the masks, assuming that
 * masks are OR-ed.
 */
char *generate_filter_or(mask_t * mask, unsigned int masknum,
			  unsigned int masklen)
{
    const unsigned int searchbuf = masklen + masknum * 6 + 8;
    unsigned int len = 0;
    char *search;
    mask_t **mfifo;		/* Mailbox FIFO queue. */
    mask_t *mf;			/* Mask returned from FIFO. */

    search = (char *)xmalloc(sizeof(char) * searchbuf);
    mfifo = (mask_t **) xmalloc(sizeof(mask_t *) * (masknum + 1));

    search[0] = 0;
    empty_fifo(mfifo);

    strncat(search, "ALL ", searchbuf - len - 1);
    len += 4;

    while (mask) {
	queue_fifo(mfifo, mask);
	mask = mask->next;

	while (mask && mask->type == MASK_TYPE_AND) {
	    queue_fifo(mfifo, mask);
	    mask = mask->next;
	}

	if (mask) {
	    if (len == 4 && search[0] == 'A')
		search[0] = len = 0;

	    strncat(search, "OR ", searchbuf - len - 1);
	    len += 3;
	}
	if (search[0] != 'A') {
	    search[len] = '(';
	    search[++len] = 0;
	}
	while ((mf = dequeue_fifo(mfifo))) {
	    strncat(search, mf->body, searchbuf - len - 1);
	    len = strlen(search);
	    search[len] = ' ';
	    search[++len] = 0;
	}

	if (strchr(search, '(')) {
	    search[len - 1] = ')';
	    search[len] = ' ';
	    search[++len] = 0;
	}
	empty_fifo(mfifo);
    }

    search[len - 1] = 0;

    xfree(mfifo);

    return search;
}


/*
 * Apply the appropriate action.
 */
int apply_action(char *mesgs, unsigned int *type, account_t * raccount,
		  char *destmbox, unsigned int *msgflags, char *args)
{
    unsigned int cnt;

    if (!*mesgs)
	return 0;

    log_info(LOG_ACTION, type);
    log_info(LOG_DESTINATION_MAILBOX, destmbox);

    cnt = count_messages(mesgs);

    switch (*type) {
    case FILTER_ACTION_DELETE:
	info("%d message%s deleted.\n", cnt, plural(cnt));
	action_delete(mesgs, args);
	break;
    case FILTER_ACTION_COPY:
	info("%d message%s copied to mailbox \"%s\".\n", cnt, plural(cnt),
	     destmbox);
	action_copy(mesgs, apply_namespace(destmbox, nsppri.prefix,
					   nsppri.delim), args);
	break;
    case FILTER_ACTION_MOVE:
	info("%d message%s moved to mailbox \"%s\".\n", cnt, plural(cnt),
	     destmbox);
	action_move(mesgs, apply_namespace(destmbox, nsppri.prefix,
					   nsppri.delim), args);
	break;
    case FILTER_ACTION_RCOPY:
	info("%d message%s copied to mailbox \"%s\" at account %s.\n", cnt,
	     plural(cnt), destmbox, raccount->key);
	action_rcopy(mesgs, raccount, destmbox, args);
	break;
    case FILTER_ACTION_RMOVE:
	info("%d message%s moved to mailbox \"%s\" at account %s.\n", cnt,
	     plural(cnt), destmbox, raccount->key);
	action_rmove(mesgs, raccount, destmbox, args);
	break;
    case FILTER_ACTION_FLAG_REPLACE:
    case FILTER_ACTION_FLAG_ADD:
    case FILTER_ACTION_FLAG_REMOVE:
	info("%d message%s flagged.\n", cnt, plural(cnt));
	action_flag(mesgs, type, msgflags, args);
	break;
    case FILTER_ACTION_LIST:
	info("%d message%s listed.\n", cnt, plural(cnt));
	action_list(mesgs, args);
	break;
    }

    if (!*args)
	log_info(LOG_WRITE, NULL);

    return 0;
}


/*
 * Delete messages and optionally list some of their headers.
 */
int action_delete(char *mesgs, char *args)
{
    char *tok, *m, *mcp;

    action_list(mesgs, args);

    m = mcp = convert_messages(mesgs);

    tok = strtok_r(m, " ", &m);
    while (tok) {
	server_response(&sockpri, imap_store(&sockpri, tok,
					     STORE_FLAG_ADD, "\\Deleted"));

	tok = strtok_r(NULL, " ", &m);
    }

    if (options & OPTION_EXPUNGE)
	server_response(&sockpri, imap_expunge(&sockpri));

    xfree(mcp);

    return 0;
}


/*
 * Copy messages to specified mailbox.
 */
int action_copy(char *mesgs, char *destmbox, char *args)
{
    int r = 0;
    char *tok = NULL, *mcp, *m;

    action_list(mesgs, args);

    m = mcp = convert_messages(mesgs);

    tok = strtok_r(m, " ", &m);
    while (tok) {
	if ((r = copy_response(&sockpri,
			       imap_copy(&sockpri, tok, destmbox))) ==
	    RESPONSE_TRYCREATE)
	    if (!server_response(&sockpri, imap_create(&sockpri, destmbox))) {
		if ((options & OPTION_SUBSCRIBE))
		    server_response(&sockpri,
				    imap_subscribe(&sockpri, destmbox));
		r = copy_response(&sockpri,
				  imap_copy(&sockpri, tok, destmbox));
	    }
	tok = strtok_r(NULL, " ", &m);
    }

    xfree(mcp);

    return r;
}


/*
 * Move messages to specified mailbox.
 */
int action_move(char *mesgs, char *destmbox, char *args)
{
    if (!action_copy(mesgs, destmbox, args))
	action_delete(mesgs, "\0");

    return 0;
}


/*
 * Copy messages to the specified mailbox of another mail server.
 */
int action_rcopy(char *mesgs, account_t * destacc, char *destmbox, char *args)
{
    int r, ta, tf;
    char *tok, *m, *mcp, *ndm;
    unsigned int n;
    char buf[RESPONSE_BUF * 2];

    if (init_connection(&sockaux, destacc->server, destacc->port,
			destacc->ssl))
	return ERROR_NETWORK;

    r = greeting_response(&sockaux);

    if (r == RESPONSE_BYE || check_capabilities(&sockaux))
	return ERROR_NETWORK;

#ifdef DEBUG
    test(&sockaux);
#endif

    if (r != RESPONSE_PREAUTH) {
	if (destacc->passwdattr == PASSWORD_NONE) {
	    printf("Enter password for %s@%s: ", destacc->username,
		   destacc->server);
	    get_password(destacc->password, PASSWORD_LEN);
	    destacc->passwdattr = PASSWORD_PLAIN;
	}
	if (login(&sockaux, destacc->username,
		  destacc->password) == RESPONSE_NO) {
	    error("imapfilter: username %s or password rejected at %s\n",
		  destacc->username, destacc->server);
	    return ERROR_NETWORK;
	}
    }
    check_namespace(&sockaux, &nspaux);

    /* apply_namespace() returns a pointer to a static buffer. */
    ndm = apply_namespace(destmbox, nspaux.prefix, nspaux.delim);

    r = check_mailbox(&sockaux, destmbox, &nspaux);

    if (r == RESPONSE_OK)
	close_mailbox(&sockaux);
    else if (r == RESPONSE_NO) {
	server_response(&sockaux, imap_create(&sockaux, ndm));
	if ((options & OPTION_SUBSCRIBE))
	    server_response(&sockaux, imap_subscribe(&sockaux, ndm));
    }
    m = mcp = xstrdup(mesgs);

    tok = strtok_r(m, " ", &m);
    while (tok) {
	fetchsize_response(&sockpri, &n,
			   imap_fetch(&sockpri, tok, "RFC822.SIZE"));

	ta = imap_append(&sockaux, ndm, n);

	fetch_response(&sockpri, 0, 1, NULL);
	tf = imap_fetch(&sockpri, tok, "RFC822.HEADER");
	do {
	    r = fetch_response(&sockpri, tf, 0, buf);
	    socket_write(&sockaux, buf);
	} while (r == RESPONSE_NONE);

	socket_write(&sockaux, "\r\n");

	fetch_response(&sockpri, 0, 1, NULL);
	tf = imap_fetch(&sockpri, tok, "BODY[TEXT]");
	do {
	    r = fetch_response(&sockpri, tf, 0, buf);
	    if (r != RESPONSE_NULLBODY)
		socket_write(&sockaux, buf);
	} while (r == RESPONSE_NONE);

	if (r != RESPONSE_NULLBODY)
	    socket_write(&sockaux, "\r\n\r\n");
	else
	    socket_write(&sockaux, "\r\n");

	append_response(&sockaux, ta);

	tok = strtok_r(NULL, " ", &m);
    }

    logout(&sockaux);

    action_list(mesgs, args);

    xfree(mcp);

    return 0;
}


/*
 * Move messages to the specified mailbox of another mail server.
 */
int action_rmove(char *mesgs, account_t * destacc, char *destmbox, char *args)
{
    if (!action_rcopy(mesgs, destacc, destmbox, args))
	action_delete(mesgs, "\0");

    return 0;
}


/*
 * Flag messages by replacing, adding or removing specified flags.
 */
int action_flag(char *mesgs, unsigned int *type, unsigned int *msgflags,
		 char *args)
{
    unsigned int t;
    char s[STORE_FLAGS_BUF];
    char *tok, *m, *mcp;

    *s = 0;

    switch (*type) {
    case FILTER_ACTION_FLAG_ADD:
	t = STORE_FLAG_ADD;
	break;
    case FILTER_ACTION_FLAG_REMOVE:
	t = STORE_FLAG_REMOVE;
	break;
    default:
	t = STORE_FLAG_REPLACE;
    }

    if ((*msgflags != MESSAGE_FLAG_NONE)) {
	if ((*msgflags & MESSAGE_FLAG_SEEN))
	    strncat(s, "\\Seen ", STORE_FLAGS_BUF - strlen(s) - 1);
	if ((*msgflags & MESSAGE_FLAG_ANSWERED))
	    strncat(s, "\\Answered ", STORE_FLAGS_BUF - strlen(s) - 1);
	if ((*msgflags & MESSAGE_FLAG_FLAGGED))
	    strncat(s, "\\Flagged ", STORE_FLAGS_BUF - strlen(s) - 1);
	if ((*msgflags & MESSAGE_FLAG_DELETED))
	    strncat(s, "\\Deleted ", STORE_FLAGS_BUF - strlen(s) - 1);
	if ((*msgflags & MESSAGE_FLAG_DRAFT))
	    strncat(s, "\\Draft", STORE_FLAGS_BUF - strlen(s) - 1);
	if ((s[strlen(s) - 1] == ' '))
	    s[strlen(s) - 1] = 0;
    }
    action_list(mesgs, args);

    m = mcp = convert_messages(mesgs);

    tok = strtok_r(m, " ", &m);
    while (tok) {
	server_response(&sockpri, imap_store(&sockpri, tok, t, s));

	tok = strtok_r(NULL, " ", &m);
    }

    if (options & OPTION_EXPUNGE)
	server_response(&sockpri, imap_expunge(&sockpri));

    xfree(mcp);

    return 0;
}

/*
 * List user selected headers of messages.
 */
int action_list(char *mesgs, char *args)
{
    int r, t;
    char *tok, *mcp, *m;
    char s[ARGS_LEN + 27];
    char hdrs[RESPONSE_BUF];

    if (!*args)
	return 0;

    m = mcp = xstrdup(mesgs);

    snprintf(s, ARGS_LEN + 27 - 1, "BODY.PEEK[HEADER.FIELDS (%s)]", args);

    tok = strtok_r(m, " ", &m);
    while (tok) {
	/* Reset internal fetch counter. */
	fetch_response(&sockpri, 0, 1, NULL);
	t = imap_fetch(&sockpri, tok, s);
	do
	    r = fetch_response(&sockpri, t, 0, hdrs);
	while (r == RESPONSE_NONE);

	if (*hdrs) {
	    if (options & OPTION_HEADERS)
		info("%s\n", hdrs);
	    log_info(LOG_WRITE, hdrs);
	} else {
	    log_info(LOG_WRITE, NULL);
	}

	tok = strtok_r(NULL, " ", &m);
    }

    xfree(mcp);

    return 0;
}


/*
 * Count how many messages matched the filter.
 */
unsigned int count_messages(char *mesgs)
{
    unsigned int cnt = 0;
    char *c = mesgs;

    while ((c = strchr(c, ' '))) {
	cnt++;
	c++;
    }

    return ++cnt;
}


/*
 * Convert messages with contiguous sequence number to the corresponding
 * number range, eg. 1 2 3 5 7 8 --> 1:3 5 7:8 and return a newly allocated
 * buffer with the results.
 */
char *convert_messages(char *mesgs)
{
    int maxlen;
    unsigned int start, end, tmp;
    char *c, *cp, *tail;

    start = end = tmp = 0;
    maxlen = strlen(mesgs) + 1;
    tail = NULL;

    c = cp = xstrdup(mesgs);

    start = (unsigned int)strtoul(mesgs, &tail, 10);
    end = start;

    do {
	if (tail) {
	    tmp = (unsigned int)strtoul(tail, &tail, 10);
	    if (!tmp)
		tail = NULL;
	}
	if (tmp == end + 1)
	    end++;
	else {
	    if (start == end) {
		xstrncpy(c, ultostr(start, 10), maxlen);
		c += strlen(c);
	    } else {
		xstrncpy(c, ultostr(start, 10), maxlen - 1);
		c += strlen(c);
		*c = ':';
		*++c = 0;
		xstrncpy(c, ultostr(end, 10), maxlen);
		c += strlen(c);
	    }

	    if (tail && c - cp < maxlen) {
		*c = ' ';
		*++c = 0;
	    }
	    start = end = tmp;
	}
    } while (tmp);

    return cp;
}
